# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""The `airbyte-cdk.cli` module provides command-line interfaces for the Airbyte CDK.

As of now, it includes the following CLI entry points:

- `airbyte-cdk`: Commands for working with connectors.
- `source-declarative-manifest`: Directly invoke the declarative manifests connector.

"""
