# Copyright (c) 2021 Airbyte, Inc., all rights reserved.
"""The destinations module provides classes for building destination connectors."""

from .destination import Destination

__all__ = [
    "Destination",
]
